@@ -1,8 +1,11 @@
 package main
 
 //go:generate go run gen.go -out ../encodeblock_amd64.s -stubs ../encodeblock_amd64.go -pkg=s2
+//go:generate gofmt -w ../encodeblock_amd64.go
+//go:generate go run cleanup.go ../encodeblock_amd64.s
 
 import (
+	"flag"
 	"fmt"
 	"math"
 	"runtime"
@@ -27,11 +30,24 @@ const (
 )
 
 func main() {
+	flag.Parse()
 	Constraint(buildtags.Not("appengine").ToConstraint())
 	Constraint(buildtags.Not("noasm").ToConstraint())
 	Constraint(buildtags.Term("gc").ToConstraint())
+	Constraint(buildtags.Not("noasm").ToConstraint())
+
+	// We need a function to add comments.
+	TEXT("_dummy_", 0, "func()")
+	Comment("#ifdef GOAMD64_v4")
+	Comment("#ifndef GOAMD64_v3")
+	Comment("#define GOAMD64_v3")
+	Comment("#endif")
+	Comment("#endif")
+	RET()
 
 	o := options{
+		bmi1:         false,
+		bmi2:         false,
 		snappy:       false,
 		outputMargin: 9,
 	}
@@ -43,12 +59,12 @@ func main() {
 
 	o.outputMargin = 6
 	o.maxSkip = 100 // Blocks can be long, limit max skipping.
-	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm", 16, 7, 7, limit14B)
-	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm4MB", 16, 7, 7, 4<<20)
+	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm", 17, 14, 7, 7, limit14B)
+	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm4MB", 17, 14, 7, 7, 4<<20)
 	o.maxSkip = 0
-	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm12B", 14, 6, 6, limit12B)
-	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm10B", 12, 5, 6, limit10B)
-	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm8B", 10, 4, 6, limit8B)
+	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm12B", 14, 12, 6, 6, limit12B)
+	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm10B", 12, 10, 5, 6, limit10B)
+	o.genEncodeBetterBlockAsm("encodeBetterBlockAsm8B", 10, 8, 4, 6, limit8B)
 
 	// Snappy compatible
 	o.snappy = true
@@ -60,12 +76,12 @@ func main() {
 	o.genEncodeBlockAsm("encodeSnappyBlockAsm8B", 8, 4, 4, limit8B)
 
 	o.maxSkip = 100
-	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm", 16, 7, 7, limit14B)
+	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm", 17, 14, 7, 7, limit14B)
 	o.maxSkip = 0
-	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm64K", 16, 7, 7, 64<<10-1)
-	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm12B", 14, 6, 6, limit12B)
-	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm10B", 12, 5, 6, limit10B)
-	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm8B", 10, 4, 6, limit8B)
+	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm64K", 16, 14, 7, 7, 64<<10-1)
+	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm12B", 14, 12, 6, 6, limit12B)
+	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm10B", 12, 10, 5, 6, limit10B)
+	o.genEncodeBetterBlockAsm("encodeSnappyBetterBlockAsm8B", 10, 8, 4, 6, limit8B)
 
 	o.snappy = false
 	o.outputMargin = 0
@@ -114,7 +130,8 @@ func assert(fn func(ok LabelRef)) {
 
 type options struct {
 	snappy       bool
-	vmbi2        bool
+	bmi1         bool
+	bmi2         bool
 	maxLen       int
 	outputMargin int // Should be at least 5.
 	maxSkip      int
@@ -424,7 +441,7 @@ func (o options) genEncodeBlockAsm(name string, tableBits, skipLog, hashBytes, m
 					JZ(LabelRef("repeat_as_copy_" + name))
 
 					// Emit as repeat...
-					o.emitRepeat("match_repeat_"+name, length, offsetVal, nil, dst, LabelRef("repeat_end_emit_"+name))
+					o.emitRepeat("match_repeat_"+name, length, offsetVal, nil, dst, LabelRef("repeat_end_emit_"+name), false)
 
 					// Emit as copy instead...
 					Label("repeat_as_copy_" + name)
@@ -710,19 +727,25 @@ func (o options) genEncodeBlockAsm(name string, tableBits, skipLog, hashBytes, m
 		RET()
 		Label("emit_remainder_ok_" + name)
 	}
-	// emitLiteral(dst[d:], src[nextEmitL:])
-	emitEnd := GP64()
-	MOVQ(lenSrcQ, emitEnd)
+	{
+		// emitLiteral(dst[d:], src[nextEmitL:])
+		emitEnd := GP64()
+		MOVQ(lenSrcQ, emitEnd)
 
-	// Emit final literals.
-	o.emitLiteralsDstP(nextEmitL, emitEnd, src, dst, "emit_remainder_"+name)
-
-	// Assert size is < limit
-	assert(func(ok LabelRef) {
-		// if dstBaseQ <  dstLimitPtrQ: ok
-		CMPQ(dst, dstLimitPtrQ)
-		JL(ok)
-	})
+		// Emit final literals.
+		// Since we may be at the end of source,
+		// we cannot have output margin.
+		x := o.outputMargin
+		o.outputMargin = 0
+		o.emitLiteralsDstP(nextEmitL, emitEnd, src, dst, "emit_remainder_"+name)
+		o.outputMargin = x
+		// Assert size is < limit
+		assert(func(ok LabelRef) {
+			// if dstBaseQ <  dstLimitPtrQ: ok
+			CMPQ(dst, dstLimitPtrQ)
+			JL(ok)
+		})
+	}
 
 	// length := start - base (ptr arithmetic)
 	length := GP64()
@@ -762,7 +785,7 @@ func maxLitOverheadFor(n int) int {
 	return 5
 }
 
-func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHashBytes, maxLen int) {
+func (o options) genEncodeBetterBlockAsm(name string, lTableBits, sTableBits, skipLog, lHashBytes, maxLen int) {
 	TEXT(name, 0, "func(dst, src []byte) int")
 	Doc(name+" encodes a non-empty src to a guaranteed-large-enough dst.",
 		fmt.Sprintf("Maximum input %d bytes.", maxLen),
@@ -774,7 +797,6 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 	}
 	var literalMaxOverhead = maxLitOverheadFor(maxLen)
 
-	var sTableBits = lTableBits - 2
 	const sHashBytes = 4
 	o.maxLen = maxLen
 
@@ -975,10 +997,34 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 			MOVL(s, sTab.Idx(hash1, 4))
 		}
 
+		longVal := GP64()
+		shortVal := GP64()
+		MOVQ(Mem{Base: src, Index: candidate, Scale: 1}, longVal)
+		MOVQ(Mem{Base: src, Index: candidateS, Scale: 1}, shortVal)
+
+		// If we have at least 8 bytes match, choose that first.
+		CMPQ(longVal, cv.As64())
+		JEQ(LabelRef("candidate_match_" + name))
+
+		CMPQ(shortVal, cv.As64())
+		JNE(LabelRef("no_short_found_" + name))
+		MOVL(candidateS.As32(), candidate.As32())
+		JMP(LabelRef("candidate_match_" + name))
+
+		Label("no_short_found_" + name)
+		MOVL(longVal.As32(), longVal.As32())
+
 		// En/disable repeat matching.
+		// Too small improvement
 		if false {
+			{
+				CMPL(repeatL, U8(0))
+				JEQ(LabelRef("no_repeat_found_" + name))
+			}
 			// Check repeat at offset checkRep
 			const checkRep = 1
+			const wantRepeatBytes = 6
+			const repeatMask = ((1 << (wantRepeatBytes * 8)) - 1) << (8 * checkRep)
 			{
 				// rep = s - repeat
 				rep := GP32()
@@ -987,10 +1033,13 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 
 				// if uint32(cv>>(checkRep*8)) == load32(src, s-repeat+checkRep) {
 				left, right := GP64(), GP64()
-				MOVL(Mem{Base: src, Index: rep, Disp: checkRep, Scale: 1}, right.As32())
+				MOVQ(Mem{Base: src, Index: rep, Disp: 0, Scale: 1}, right.As64())
 				MOVQ(cv, left)
-				SHRQ(U8(checkRep*8), left)
-				CMPL(left.As32(), right.As32())
+				tmp := GP64()
+				MOVQ(U64(repeatMask), tmp)
+				ANDQ(tmp, left)
+				ANDQ(tmp, right)
+				CMPQ(left.As64(), right.As64())
 				// BAIL, no repeat.
 				JNE(LabelRef("no_repeat_found_" + name))
 			}
@@ -1034,7 +1083,7 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 			// Extend forward
 			{
 				// s += 4 + checkRep
-				ADDL(U8(4+checkRep), s)
+				ADDL(U8(wantRepeatBytes+checkRep), s)
 
 				if true {
 					// candidate := s - repeat + 4 + checkRep
@@ -1074,18 +1123,8 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 				offsetVal := GP32()
 				MOVL(repeatL, offsetVal)
 
-				if !o.snappy {
-					// if nextEmit == 0 {do copy instead...}
-					TESTL(nextEmit, nextEmit)
-					JZ(LabelRef("repeat_as_copy_" + name))
-
-					// Emit as repeat...
-					o.emitRepeat("match_repeat_"+name, length, offsetVal, nil, dst, LabelRef("repeat_end_emit_"+name))
-
-					// Emit as copy instead...
-					Label("repeat_as_copy_" + name)
-				}
-				o.emitCopy("repeat_as_copy_"+name, length, offsetVal, nil, dst, LabelRef("repeat_end_emit_"+name))
+				// Emit as repeat...
+				o.emitRepeat("match_repeat_"+name, length, offsetVal, nil, dst, LabelRef("repeat_end_emit_"+name), false)
 
 				Label("repeat_end_emit_" + name)
 				// Store new dst and nextEmit
@@ -1122,11 +1161,11 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 				JG(ok)
 			})
 
-			CMPL(Mem{Base: src, Index: candidate, Scale: 1}, cv.As32())
+			CMPL(longVal.As32(), cv.As32())
 			JEQ(LabelRef("candidate_match_" + name))
 
 			//if uint32(cv) == load32(src, candidateS)
-			CMPL(Mem{Base: src, Index: candidateS, Scale: 1}, cv.As32())
+			CMPL(shortVal.As32(), cv.As32())
 			JEQ(LabelRef("candidateS_match_" + name))
 
 			// No match found, next loop
@@ -1291,7 +1330,7 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 				// length += 4
 				ADDL(U8(4), length.As32())
 				MOVL(s, nextEmitL) // nextEmit = s
-				o.emitRepeat("match_nolit_repeat_"+name, length, offset, nil, dst, LabelRef("match_nolit_emitcopy_end_"+name))
+				o.emitRepeat("match_nolit_repeat_"+name, length, offset, nil, dst, LabelRef("match_nolit_emitcopy_end_"+name), false)
 			}
 		}
 		Label("match_nolit_emitcopy_end_" + name)
@@ -1315,11 +1354,57 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 		}
 	}
 	Label("match_nolit_dst_ok_" + name)
-	// cv must be set to value at base+1 before arriving here
 	if true {
 		lHasher := hashN(lHashBytes, lTableBits)
 		sHasher := hashN(sHashBytes, sTableBits)
 
+		index0, index1 := GP64(), GP64()
+		// index0 := base + 1
+		LEAQ(Mem{Base: base, Disp: 1}, index0)
+		// index1 := s - 2
+		LEAQ(Mem{Base: s, Disp: -2}, index1)
+		hash0l, hash0s, hash1l, hash1s := GP64(), GP64(), GP64(), GP64()
+		MOVQ(Mem{Base: src, Index: index0, Scale: 1, Disp: 0}, hash0l)
+		MOVQ(Mem{Base: src, Index: index0, Scale: 1, Disp: 1}, hash0s)
+		MOVQ(Mem{Base: src, Index: index1, Scale: 1, Disp: 0}, hash1l)
+		MOVQ(Mem{Base: src, Index: index1, Scale: 1, Disp: 1}, hash1s)
+
+		lHasher.hash(hash0l)
+		sHasher.hash(hash0s)
+		lHasher.hash(hash1l)
+		sHasher.hash(hash1s)
+
+		plusone0, plusone1 := GP64(), GP64()
+		LEAQ(Mem{Base: index0, Disp: 1}, plusone0)
+		LEAQ(Mem{Base: index1, Disp: 1}, plusone1)
+		MOVL(index0.As32(), lTab.Idx(hash0l, 4))
+		MOVL(index1.As32(), lTab.Idx(hash1l, 4))
+		MOVL(plusone0.As32(), sTab.Idx(hash0s, 4))
+		MOVL(plusone1.As32(), sTab.Idx(hash1s, 4))
+
+		ADDQ(U8(1), index0)
+		SUBQ(U8(1), index1)
+
+		Label("index_loop_" + name)
+		CMPQ(index0, index1)
+		JAE(LabelRef("search_loop_" + name))
+		hash0l, hash1l = GP64(), GP64()
+		MOVQ(Mem{Base: src, Index: index0, Scale: 1, Disp: 0}, hash0l)
+		MOVQ(Mem{Base: src, Index: index1, Scale: 1, Disp: 0}, hash1l)
+
+		lHasher.hash(hash0l)
+		lHasher.hash(hash1l)
+
+		MOVL(index0.As32(), lTab.Idx(hash0l, 4))
+		MOVL(index1.As32(), lTab.Idx(hash1l, 4))
+
+		ADDQ(U8(2), index0)
+		SUBQ(U8(2), index1)
+		JMP(LabelRef("index_loop_" + name))
+	} else {
+		lHasher := hashN(lHashBytes, lTableBits)
+		sHasher := hashN(sHashBytes, sTableBits)
+
 		// Index base+1 long, base+2 short...
 		cv := GP64()
 		INCL(base)
@@ -1389,8 +1474,8 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 		MOVL(sm2, lTab.Idx(hash0, 4))
 		MOVL(sm1, sTab.Idx(hash1, 4))
 		MOVL(sm1, lTab.Idx(hash3, 4))
+		JMP(LabelRef("search_loop_" + name))
 	}
-	JMP(LabelRef("search_loop_" + name))
 
 	Label("emit_remainder_" + name)
 	// Bail if we exceed the maximum size.
@@ -1420,7 +1505,12 @@ func (o options) genEncodeBetterBlockAsm(name string, lTableBits, skipLog, lHash
 	MOVQ(lenSrcQ, emitEnd)
 
 	// Emit final literals.
+	// Since we may be at the end of source,
+	// we cannot have output margin.
+	x := o.outputMargin
+	o.outputMargin = 0
 	o.emitLiteralsDstP(nextEmitL, emitEnd, src, dst, "emit_remainder_"+name)
+	o.outputMargin = x
 
 	// Assert size is < limit
 	assert(func(ok LabelRef) {
@@ -1550,8 +1640,12 @@ func (h hashGen) hash(val reg.GPVirtual) {
 	if h.bytes < 8 {
 		SHLQ(U8(64-8*h.bytes), val)
 	}
+	//  329 AMD64               :IMUL r64, r64                         L:   0.86ns=  3.0c  T:   0.29ns=  1.00c
+	// 2020 BMI2                :MULX r64, r64, r64                    L:   1.14ns=  4.0c  T:   0.29ns=  1.00c
 	IMULQ(h.mulreg, val)
 	// Move value to bottom
+	// 2032 BMI2                :SHRX r64, r64, r64                    L:   0.29ns=  1.0c  T:   0.12ns=  0.42c
+	//  236 AMD64               :SHR r64, imm8                         L:   0.29ns=  1.0c  T:   0.13ns=  0.46c
 	SHRQ(U8(64-h.tablebits), val)
 }
 
@@ -1742,7 +1836,7 @@ func (o options) genEmitRepeat() {
 	Load(Param("dst").Base(), dstBase)
 	Load(Param("offset"), offset)
 	Load(Param("length"), length)
-	o.emitRepeat("standalone", length, offset, retval, dstBase, LabelRef("gen_emit_repeat_end"))
+	o.emitRepeat("standalone", length, offset, retval, dstBase, LabelRef("gen_emit_repeat_end"), false)
 	Label("gen_emit_repeat_end")
 	Store(retval, ReturnIndex(0))
 	RET()
@@ -1754,13 +1848,18 @@ func (o options) genEmitRepeat() {
 // retval can be nil.
 // Will jump to end label when finished.
 // Uses 1 GP register.
-func (o options) emitRepeat(name string, length, offset, retval, dstBase reg.GPVirtual, end LabelRef) {
+// longer indicates we know match will be > 12
+func (o options) emitRepeat(name string, length reg.GPVirtual, offset reg.GPVirtual, retval reg.GPVirtual, dstBase reg.GPVirtual, end LabelRef, longer bool) {
 	Comment("emitRepeat")
+	if longer {
+		// Skip initial length tests
+		LEAL(Mem{Base: length, Disp: -4}, length.As32()) // length -= 4
+		JMP(LabelRef("cant_repeat_two_offset_" + name))
+	}
 	Label("emit_repeat_again_" + name)
 	tmp := GP32()
-	MOVL(length.As32(), tmp) // Copy length
-	// length -= 4
-	LEAL(Mem{Base: length, Disp: -4}, length.As32())
+	MOVL(length.As32(), tmp)                         // Copy length
+	LEAL(Mem{Base: length, Disp: -4}, length.As32()) // length -= 4
 
 	// if length <= 4 (use copied value)
 	CMPL(tmp.As32(), U8(8))
@@ -1982,7 +2081,7 @@ func (o options) emitCopy(name string, length, offset, retval, dstBase reg.GPVir
 		//	return 5 + emitRepeat(dst[5:], offset, length)
 		// Inline call to emitRepeat. Will jump to end
 		if !o.snappy {
-			o.emitRepeat(name+"_emit_copy", length, offset, retval, dstBase, end)
+			o.emitRepeat(name+"_emit_copy", length, offset, retval, dstBase, end, false)
 		}
 		JMP(LabelRef("four_bytes_loop_back_" + name))
 
@@ -2018,6 +2117,37 @@ func (o options) emitCopy(name string, length, offset, retval, dstBase reg.GPVir
 	//if length > 64 {
 	CMPL(length.As32(), U8(64))
 	JLE(LabelRef("two_byte_offset_short_" + name))
+
+	// if offset < 2048 {
+	if !o.snappy {
+		CMPL(offset.As32(), U32(2048))
+		JAE(LabelRef("long_offset_short_" + name))
+		// Emit a length 8 copy, encoded as 2 bytes.
+		// Emit remaining as repeat value (minimum 4 bytes).
+		// dst[1] = uint8(offset)
+		// dst[0] = uint8(offset>>8)<<5 | uint8(8-4)<<2 | tagCopy1
+		tmp := GP64()
+		MOVL(U32(tagCopy1), tmp.As32())
+		// Use scale and displacement to shift and subtract values from length.
+		LEAL(Mem{Base: tmp, Disp: (8 - 4) << 2}, tmp.As32())
+		MOVB(offset.As8(), Mem{Base: dstBase, Disp: 1}) // Store offset lower byte
+		tmp2 := GP64()
+		MOVL(offset.As32(), tmp2.As32())
+		SHRL(U8(8), tmp2.As32())     // Remove lower
+		SHLL(U8(5), tmp2.As32())     // Shift back up
+		ORL(tmp2.As32(), tmp.As32()) // OR result
+		MOVB(tmp.As8(), Mem{Base: dstBase, Disp: 0})
+		if retval != nil {
+			ADDQ(U8(2), retval) // i += 2
+		}
+		ADDQ(U8(2), dstBase)       // dst += 2
+		SUBL(U8(8), length.As32()) // length  -= 8
+		// emitRepeat(dst[2:], offset, length)
+		o.emitRepeat(name+"_emit_copy_short_2b", length, offset, retval, dstBase, end, true)
+
+		Label("long_offset_short_" + name)
+	}
+
 	// Emit a length 60 copy, encoded as 3 bytes.
 	// Emit remaining as repeat value (minimum 4 bytes).
 	//	dst[2] = uint8(offset >> 8)
@@ -2037,7 +2167,7 @@ func (o options) emitCopy(name string, length, offset, retval, dstBase reg.GPVir
 	}
 	// Inline call to emitRepeat. Will jump to end
 	if !o.snappy {
-		o.emitRepeat(name+"_emit_copy_short", length, offset, retval, dstBase, end)
+		o.emitRepeat(name+"_emit_copy_short", length, offset, retval, dstBase, end, false)
 	}
 	JMP(LabelRef("two_byte_offset_" + name))
 
@@ -2482,7 +2612,7 @@ func (o options) matchLen(name string, a, b, len reg.GPVirtual, end LabelRef) re
 	XORL(matched, matched)
 
 	CMPL(len.As32(), U8(8))
-	JL(LabelRef("matchlen_single_" + name))
+	JL(LabelRef("matchlen_match4_" + name))
 
 	Label("matchlen_loopback_" + name)
 	MOVQ(Mem{Base: a, Index: matched, Scale: 1}, tmp)
@@ -2490,7 +2620,15 @@ func (o options) matchLen(name string, a, b, len reg.GPVirtual, end LabelRef) re
 	TESTQ(tmp, tmp)
 	JZ(LabelRef("matchlen_loop_" + name))
 	// Not all match.
+
+	Comment("#ifdef GOAMD64_v3")
+	// 2016 BMI                 :TZCNT r64, r64                        L:   0.57ns=  2.0c  T:   0.29ns=  1.00c
+	//  315 AMD64               :BSF r64, r64                          L:   0.88ns=  3.1c  T:   0.86ns=  3.00c
+	TZCNTQ(tmp, tmp)
+	Comment("#else")
 	BSFQ(tmp, tmp)
+	Comment("#endif")
+
 	SARQ(U8(3), tmp)
 	LEAL(Mem{Base: matched, Index: tmp, Scale: 1}, matched)
 	JMP(end)
@@ -2501,18 +2639,37 @@ func (o options) matchLen(name string, a, b, len reg.GPVirtual, end LabelRef) re
 	LEAL(Mem{Base: matched, Disp: 8}, matched)
 	CMPL(len.As32(), U8(8))
 	JGE(LabelRef("matchlen_loopback_" + name))
+	JZ(end)
 
 	// Less than 8 bytes left.
-	Label("matchlen_single_" + name)
-	TESTL(len.As32(), len.As32())
-	JZ(end)
-	Label("matchlen_single_loopback_" + name)
+	// Test 4 bytes...
+	Label("matchlen_match4_" + name)
+	CMPL(len.As32(), U8(4))
+	JL(LabelRef("matchlen_match2_" + name))
+	MOVL(Mem{Base: a, Index: matched, Scale: 1}, tmp.As32())
+	CMPL(Mem{Base: b, Index: matched, Scale: 1}, tmp.As32())
+	JNE(LabelRef("matchlen_match2_" + name))
+	SUBL(U8(4), len.As32())
+	LEAL(Mem{Base: matched, Disp: 4}, matched)
+
+	// Test 2 bytes...
+	Label("matchlen_match2_" + name)
+	CMPL(len.As32(), U8(2))
+	JL(LabelRef("matchlen_match1_" + name))
+	MOVW(Mem{Base: a, Index: matched, Scale: 1}, tmp.As16())
+	CMPW(Mem{Base: b, Index: matched, Scale: 1}, tmp.As16())
+	JNE(LabelRef("matchlen_match1_" + name))
+	SUBL(U8(2), len.As32())
+	LEAL(Mem{Base: matched, Disp: 2}, matched)
+
+	// Test 1 byte...
+	Label("matchlen_match1_" + name)
+	CMPL(len.As32(), U8(1))
+	JL(end)
 	MOVB(Mem{Base: a, Index: matched, Scale: 1}, tmp.As8())
 	CMPB(Mem{Base: b, Index: matched, Scale: 1}, tmp.As8())
 	JNE(end)
 	LEAL(Mem{Base: matched, Disp: 1}, matched)
-	DECL(len.As32())
-	JNZ(LabelRef("matchlen_single_loopback_" + name))
 	JMP(end)
 	return matched
 }
